import itertools
import re
from collections import defaultdict
from typing import Generator, Iterable

from .motif_classes import MotifClasses
from ..dynamic_constants import DynamicConstants


class Position(dict):
    def __init__(self, position: int, sequences: Iterable, variants_flattened: list, variant_dict: defaultdict,
                 entropy: float = None, variant_data: bool = False):
        """
            Data structure for kmer positions

            Describes the data structure for kmer positions within the alignment

            :param position: A zero-based index for position number
            :param sequences: A list of type Variant containing variants seen at current position
            :param variants_flattened: A flattened list of all variants. Note: Only used for processing
            :param variant_dict: A default dictionary of lists to store the sequence idx for later getting description data
            :param entropy: The Shannon entropy at the current kmer position
            :param variant_data: Whether to derive variant data from the sequence headers.

            :type position: int
            :type sequences: Generator
            :type variants_flattened: list
            :type variant_dict: defaultdict
            :type entropy: float
        """

        self.position = position
        self.entropy = entropy
        self.variants_flattened = variants_flattened
        self.supports = len(self.variants_flattened)
        self.sequences = self._motif_classify(sequences)
        self.variants = len(self.sequences)
        self.kmertypes = [sequence.sequence for sequence in self.sequences]

        self._set_desc_data(variant_dict) if variant_data else None

    def _set_desc_data(self, variant_dict):
        """
            Sets the data from the header as attributes in the variant object.

            :param variant_dict: A defaultdict of list containing idx of sequences
            :type variant_dict: defaultdict
        """

        expr = re.compile(DynamicConstants.HEADER_REGEX)

        for dict_variant, variant in itertools.product(variant_dict, self.sequences):
            if dict_variant != variant.sequence:
                continue
            idx = variant_dict[dict_variant]

            for ids in idx:
                header = DynamicConstants.SEQ_DESCRIPTIONS[ids]

                mapping = [x.groupdict() for x in expr.finditer(header)][0]

                for key, value in mapping.items():
                    current_value = variant.get(key)

                    if not current_value:
                        variant[key] = [value]
                        continue

                    variant[key] = current_value + [value]

    def __setattr__(self, key, value):
        # This really makes me sad, find a way around this mess TODO: fix this please, also in Variant
        dict.__setitem__(self, key, value)

    def __getattr__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            raise AttributeError

    def __getstate__(self):
        state = self.copy()
        del state['variants_flattened']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

    @classmethod
    def _motif_classify(cls, variants: Iterable) -> list:
        """
            Given a list of Variant objects sort by the count of each, and  classify each variant into motif classes

            :param variants: A list of Variant objects
            :type variants: Generator

            :return: A list containing Variant objects sorted by count and each variant classified into motif classes
        """

        variants = sorted(variants, key=lambda x: x.count, reverse=True)

        for idx, variant in enumerate(variants):
            if idx == 0:
                variants[idx].motif_short = 'I'
                variants[idx].motif_long = MotifClasses.I
            elif idx == 1:
                variants[idx].motif_short = 'Ma'
                variants[idx].motif_long = MotifClasses.Ma
            elif variants[idx].count > 1:
                variants[idx].motif_short = 'Mi'
                variants[idx].motif_long = MotifClasses.Mi
            else:
                variants[idx].motif_short = 'U'
                variants[idx].motif_long = MotifClasses.U

        return variants



import math
import random
from .datatypes import VariantCounter
from scipy.stats import stats


class NormalizedEntropy(object):
    def __init__(self, max_samples, iterations, positions: list):
        self.max_samples = max_samples
        self.iterations = iterations
        self.positions = positions

    def run(self):
        for position in self.positions:
            if not position.supports > 0:
                continue

            normalized_entropy = self._calc_normalized_entropy(position.variants_flattened)
            position.entropy = normalized_entropy

            yield position

    def _calc_normalized_entropy(self, flattened_variants):
        flattened_variants = list(flattened_variants)
        random.shuffle(flattened_variants)

        iteration_data = []

        for iteration in range(self.iterations):
            sample_count = random.randint(1, self.max_samples)
            total_iteration_entropy = 0
            samples = VariantCounter()

            for sample in range(sample_count):
                random_variant = random.choice(flattened_variants)
                samples.observe(random_variant)

            for sample_value in samples.results('values'):
                variant_entropy = self._entropy_calculation(sample_value, sample_count)
                total_iteration_entropy += variant_entropy

            iteration_data.append((total_iteration_entropy * -1, 1.0 / float(sample_count)))

        slope, intercept, r_value, p_value, std_err = stats.linregress([y for x, y in iteration_data],
                                                                       [x for x, y in iteration_data])

        return float(intercept)

    @classmethod
    def _entropy_calculation(cls, sample_value, sample_count):
        entropy = (float(sample_value) / float(sample_count)) * \
                  (math.log2(float(sample_value) / float(sample_count)))
        return entropy



import os
from collections import Counter
from itertools import islice
from typing import List, Union, Iterable

import jsonpickle
from Bio.SeqIO import parse, SeqRecord

from .core import HeaderDecode
from .entropy import NormalizedEntropy
from .datatypes import Position, Variant, VariantDict
from .dynamic_constants import DynamicConstants
from .errorhandlers import SequenceFileNotFound, HeaderDecodeError


class Hunana(object):
    DISALLOWED_CHARS = {'-', 'X', 'B', 'J', 'Z', 'O', 'U'}

    def __init__(self, seq_path: str, kmer_len: int = 9, header_decode: bool = False, json_result: bool = False,
                 max_samples: int = 10000, iterations: int = 10, header_format: str = None):
        """
            The Hunana algorithm returns a list of Position objects each corresponding to a kmer position.

            :param seq_path: The absolute path to the MSA file in FASTA format.
            :param kmer_len: The length of the kmers to generate (default:  9).
            :param header_decode: Whether to use FASTA headers to derive kmer information (default: False).
            :param json_result: Whether the results should be returned in json format (default: False).
            :param max_samples: The maximum number of samples to use when calculating entropy (default: 10000).
            :param iterations: The maximum number of iterations to use when calculating entropy (default: 10).
            :param header_format: The format of the header (ex: (id)|(species)|(country))

            :type seq_path str
            :type kmer_len: str
            :type header_decode: bool
            :type max_samples: int
            :type iterations: int
        """

        self.seqs = self._get_seqs_from_path(seq_path)
        self.kmer_len = kmer_len
        self.header_decode = header_decode
        self.json_result = json_result
        self.max_samples = max_samples
        self.iterations = iterations

        if header_decode:
            self._prepare_header_decode(self.seqs, header_format)

    @classmethod
    def _prepare_header_decode(cls, seqs: List[SeqRecord], header_format: str):
        if not header_format:
            raise HeaderDecodeError(header_format)

        DynamicConstants.SEQ_DESCRIPTIONS = [x.description for x in seqs]

        if not HeaderDecode.set_header_regex(header_format):
            raise HeaderDecodeError(header_format)

    @classmethod
    def _create_variant_dict(cls, kmers: zip) -> List[VariantDict]:
        """
            Returns a list of VariantDict (a dictionary of lists).
            TODO: Been a while, forgot what this does, try to understand again later.

            :param kmers: A zip object containing kmers
            :type kmers: zip

            :return: A list of VariantDict.
        """

        positions = []

        for kmer in kmers:
            unique = VariantDict(list)
            # Here there's no need to enumerate, you can just append any value because you are essentialy
            # only using len to get the number of elements. There has to be a more elegant way to achive the same effect
            for a, b in enumerate(kmer):
                unique[b].append(a)
            unique.pop('illegal-char', None)

            positions.append(unique)
        return positions

    @classmethod
    def _get_seqs_from_path(cls, seq_path: str) -> List[SeqRecord]:
        """
            Checks if the FASTA filepath is valid and file is present, if so, returns a list of SeqRecords for each
            sequence.

            :param seq_path: The path to the FASTA file.
            :type seq_path: str

            :return: A list of SeqRecords for each sequence in the FASTA file.
        """

        if not os.path.isfile(seq_path):
            raise SequenceFileNotFound(seq_path)

        return list(parse(seq_path, 'fasta'))

    def _kmers(self) -> zip:
        """
            Matches each kmer position of all sequences together (ex: all first kmer positions grouped together.)

            :returns: A Zip object containing all k-mers.
        """

        kmers_seqs = (self._window(seq.seq) for seq in self.seqs)
        return zip(*kmers_seqs)

    def _window(self, seq: str) -> Iterable:
        """
            Generates kmers of each sequence in it's respective element (ex: all kmers of sequence 1 in the first
            element of the generator).

            :param seq: The sequence to generate kmers out of.
            :type seq: str

            return: A Generator object containing kmers
        """

        it = iter(seq)
        result = tuple(islice(it, self.kmer_len))

        if len(result) == self.kmer_len:
            if all(ele not in result for ele in self.DISALLOWED_CHARS):
                yield ''.join(result)
            else:
                yield 'illegal-char'

        for elem in it:
            result = result[1:] + (elem,)
            if all(ele not in result for ele in self.DISALLOWED_CHARS):
                yield ''.join(result)
            else:
                yield 'illegal-char'

    def _create_position_objects(self, counters: Iterable[Counter],
                                 variant_dicts: List[VariantDict]) -> Iterable:
        """
            Create kmer position objects with entropy, position and variant properties

            :return A Generator object for Position objects for each kmer position
        """

        for idx, counter in enumerate(counters, start=1):
            yield Position(
                position=idx,
                sequences=self._create_variant_objects(idx, counter),
                variants_flattened=list(counter.elements()),
                variant_dict=variant_dicts[idx - 1],
                variant_data=self.header_decode
            )

    def _create_variant_objects(self, idx: int, variant_counter: Counter) -> Iterable[Variant]:
        """
            Creates variant objects to be inserted to Position objects later.

            :param idx: The position number.
            :param variant_counter: The counter containing the variants.

            :type idx: int
            :type variant_counter: Counter

            :return: A Generator of Variant objects.
        """

        num_variants = sum(variant_counter.values())

        variant_objects = (
            Variant(idx, seq, count, self._calc_conservation(count, num_variants))
            for seq, count
            in variant_counter.items()
        )
        return variant_objects

    @classmethod
    def _calc_conservation(cls, variant_hits: int, total_hits: int) -> float:
        """
            Calculates the conservation for a given variant.

            :param variant_hits: The number of times the variant was seen in the sequences at this particular position.
            :param total_hits: The total number of variants.

            :type variant_hits: int
            :type total_hits: int

            :return: The conservation of this particular variant.
        """

        conservation = (float(variant_hits) * 100) / float(total_hits)
        return float(conservation)

    def run(self) -> Union[str, List[Position]]:
        """
            Run the pipeline.

            :return: Returns either json results or a list of Position objects.
        """

        kmers = self._kmers()
        variant_dicts = self._create_variant_dict(kmers)
        variant_counters = (x.get_counter() for x in variant_dicts)
        kmer_position_objects = self._create_position_objects(variant_counters, variant_dicts)
        entropy = NormalizedEntropy(self.max_samples, self.iterations, list(kmer_position_objects)).run()
        results = list(entropy)

        if self.json_result:
            return jsonpickle.encode(results, indent=2, unpicklable=False)
        else:
            return results



class SequenceFileNotFound(Exception):
    def __init__(self, seq_path: str):
        """
            Error raised when the FASTA file path is invalid, or the file is not found.
            
            :param seq_path: The path to the FASTA file.
            :type seq_path: str 
        """

        msg = f'\nThe FASTA file was not found at the following location:\n\t{seq_path}\nAdvice: \n\tPlease check ' \
              f'whether the path is valid, or contains any errors.'

        super(SequenceFileNotFound, self).__init__(msg)


class HeaderDecodeError(Exception):
    def __init__(self, header_format: str):
        """
            Error raised when the correct regex pattern could not be generated.

            :param header_format: The format of the FASTA header.
            :type header_format: str
        """

        msg = f'\nFailed to generate correct regex expression using the header format:\n\t{header_format}\nAdvice:\n\t' \
              f'Make sure the header format you provided accurately reflects the format of all the FASTA headers in ' \
              f'your FASTA file.'

        super(HeaderDecodeError, self).__init__(msg)



import argparse
import sys

from hunana import Hunana


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input', help='Absolute path to the aligned sequences file in FASTA format.',
                        required=True)
    parser.add_argument('-o', '--output', help='Absolute path to the output file.', required=False)
    parser.add_argument('-l', '--length', help='The k-mer length (default: 9).', type=int, default=9,
                        choices=range(1, 15),
                        required=False)
    parser.add_argument('-s', '--samples', help='Max number of samples use when calculating entropy (default: 10000).',
                        type=int, default=10000, required=False)
    parser.add_argument('-it', '--iterations',
                        help='Max number of iterations used when calculating entropy (default: 10).',
                        type=int, default=10, required=False)
    parser.add_argument('-he', '--header', help="Should the header data be used to derive the details for variants?.",
                        action='store_true', required=False)
    parser.add_argument('-f', '--format',
                        help="The format of the header. Ex: (id)|(species)|(country). Each item enclosed "
                             "in brackets with any character used as separator.", required=False,
                        default=None)

    arguments = parser.parse_args()

    if arguments.header:
        if not arguments.format:
            parser.error(
                "The header format (-f/--format) needs to be provided if header parsing (-he/--header) parsing "
                "is enabled")
            sys.exit(6)

    if not arguments.input:
        parser.error('Error: The input path is not provided.')
        sys.exit(2)

    try:
        results = Hunana(arguments.input, arguments.length, arguments.header, True, arguments.samples,
                         arguments.iterations, arguments.format).run()
    except Exception:
        parser.error(f'Exception while calculating kmers for sequences file {arguments.input}')
        sys.exit(5)

    if not arguments.output:
        print(results)
        sys.exit(0)

    with open(arguments.output, 'w') as output:
        output.write(results)

    print(f'Results successfully saved at {arguments.output}')
    sys.exit(0)


if __name__ == '__main__':
    main()
